/* 
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2004-2007 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <sstream>

#include <odbc++/resultsetmetadata.h>

#include "data_selector.h"

using namespace TkResolver;
using namespace Db::Ui;


NoSelection:: NoSelection()
  : std::runtime_error("No selection")
{
}



DataSelector:: DataSelector(Db::Database *database,
			    const String &table_name,
			    const StringList &selection_columns)
  : _database(database),
    _table_name(table_name),
    _selection_columns(selection_columns)
{
  DEBUG_OUT << "Initialised DataSelector" << std::endl;
}



void 
DataSelector:: set_columns(const StringList &display_columns)
{
  DEBUG_OUT << "Setting columns..." << std::endl;

  
  DEBUG_OUT << "Merging name list..." << std::endl;
  StringList columns( _selection_columns.merge(display_columns) );
  DEBUG_OUT << "...merged name list" << std::endl;
  
  Gtk::TreeModelColumnRecord column_record;
  Gtk::TreeModelColumnBase * column;
  
  DEBUG_OUT << "Iterating over column names..." << std::endl;
  for (StringList::const_iterator iter(columns.begin());
       iter != columns.end();
       ++iter)
    {
      DEBUG_OUT << "Creating column model..." << std::endl;
      column = create_model_column(*iter);
      DEBUG_OUT << "...created column model" << std::endl;
      
      DEBUG_OUT << "Storing column model..." << std::endl;
      _store_columns.push_back(TreeModelColumnList::value_type(*iter, column));
      DEBUG_OUT << "...stored column model" << std::endl;
      
      DEBUG_OUT << "Adding column model to record..." << std::endl;
      column_record.add(*column);
      DEBUG_OUT << "...Added column model to record" << std::endl;
    }
  DEBUG_OUT << "...iterated over column names" << std::endl;
  
  DEBUG_OUT << "Creating list store..." << std::endl;
  _store = Gtk::ListStore::create(column_record);
  if(!_store)
    throw std::runtime_error("could not create store for data selector");
  DEBUG_OUT << "...created list store" << std::endl;

  DEBUG_OUT << "Refreshing..." << std::endl;
  refresh();
  DEBUG_OUT << "...refreshed" << std::endl;
  
  DEBUG_OUT << "...set columns" << std::endl;
}

Db::Type
DataSelector:: get_column_type(const String &column_name) const
{
  return _database->get_column_type(_table_name, column_name);
}

Db::Type
DataSelector:: get_column_type(int column_index) const
{
  TreeModelColumnList::const_iterator iter( _store_columns.begin() );
  for (int i = 0; i < column_index; ++iter, ++i);
  
  return get_column_type(iter->first);
}


Gtk::TreeModelColumnBase *
DataSelector:: create_model_column(const String &name) const
{
  Db::Type type(get_column_type(name));
  
  using namespace odbc;
  switch(type)
    {
    case Types::VARCHAR:
    case Types::LONGVARCHAR:
      DEBUG_OUT << "string column" << std::endl;
      return new Gtk::TreeModelColumn<Glib::ustring>;
    case Types::CHAR:
      DEBUG_OUT << "char column" << std::endl;
      return new Gtk::TreeModelColumn<char>;
    case Types::BIT:
      DEBUG_OUT << "bool column" << std::endl;
      return new Gtk::TreeModelColumn<bool>;
    case Types::INTEGER:
      DEBUG_OUT << "int column" << std::endl;
      return new Gtk::TreeModelColumn<int>;
    case Types::BIGINT:
      DEBUG_OUT << "Id column" << std::endl;
      return new Gtk::TreeModelColumn<Db::Id>;
    case Types::SMALLINT:
    case Types::TINYINT:
      DEBUG_OUT << "short column" << std::endl;
      return new Gtk::TreeModelColumn<short>;
    case Types::DECIMAL:
    case Types::NUMERIC:
      DEBUG_OUT << "long double column" << std::endl;
      return new Gtk::TreeModelColumn<long double>;
    case Types::DOUBLE:
    case Types::REAL:
      DEBUG_OUT << "double column" << std::endl;
      return new Gtk::TreeModelColumn<double>;
    case Types::FLOAT:
      DEBUG_OUT << "float column" << std::endl;
      return new Gtk::TreeModelColumn<float>;
    case Types::DATE:
      DEBUG_OUT << "date column" << std::endl;
      return new Gtk::TreeModelColumn<Db::Date>;
    case Types::TIMESTAMP:
      DEBUG_OUT << "datetime column" << std::endl;
      return new Gtk::TreeModelColumn<Db::DateTime>;
    default:
      throw std::runtime_error(std::string("unknown data type (") +
			       static_cast<std::string>(String(type)) +
			       ") for data column '" +
			       static_cast<std::string>(name) +
			       "'");
    }
}


Gtk::CellRenderer *
DataSelector:: create_cell(const String &column_name) const
{
  Type type(get_column_type(column_name));
  
  using namespace odbc;
  switch(type)
    {
    case Types::VARCHAR:
    case Types::LONGVARCHAR:
    case Types::CHAR:
    case Types::INTEGER:
    case Types::BIGINT:
    case Types::SMALLINT:
    case Types::TINYINT:
    case Types::DECIMAL:
    case Types::NUMERIC:
    case Types::DOUBLE:
    case Types::REAL:
    case Types::FLOAT:
    case Types::DATE:
    case Types::TIMESTAMP:
      return Gtk::manage(new Gtk::CellRendererText);
    case Types::BIT:
      return Gtk::manage(new Gtk::CellRendererToggle);
    default:
      throw std::runtime_error(std::string("unknown data type (") +
			       static_cast<std::string>(String(type)) +
			       ") for data column '" +
			       static_cast<std::string>(column_name) +
			       "'");
    }
}

String
DataSelector:: get_column_view_attribute(const String &column_name) const
{
  Type type(get_column_type(column_name));

  using namespace odbc;
  switch(type)
    {
    case Types::VARCHAR:
    case Types::LONGVARCHAR:
    case Types::CHAR:
    case Types::INTEGER:
    case Types::BIGINT:
    case Types::SMALLINT:
    case Types::TINYINT:
    case Types::DECIMAL:
    case Types::NUMERIC:
    case Types::DOUBLE:
    case Types::REAL:
    case Types::FLOAT:
    case Types::DATE:
    case Types::TIMESTAMP:
      return "text";
    case Types::BIT:
      return "active";
    default:
      throw std::runtime_error(std::string("unknown data type (") +
			       static_cast<std::string>(String(type)) +
			       ") for data column '" +
			       static_cast<std::string>(column_name) +
			       "'");
    }
}





TreeModelColumnList::const_iterator
DataSelector:: get_store_column(const String &name) const
{
  for (TreeModelColumnList::const_iterator iter(_store_columns.begin());
       iter != _store_columns.end();
       ++iter)
    {
      if(name == iter->first)
	{
	  return iter;
	}
    }

  throw std::logic_error(std::string("unknown column ") + static_cast<std::string>(name));
}


Gtk::TreeModelColumnBase *
DataSelector:: get_model_column(const String &name) const
{
  return get_store_column(name)->second;
}


unsigned
DataSelector:: get_column_index(const String &name) const
{
  return std::distance(_store_columns.begin(), get_store_column(name));
}

template<typename CV, typename V>
void
set_column_value_a(Gtk::TreeModel::iterator &iter, int column_index, const V &value)
{
  CV column_value(value);
  iter->set_value(column_index - 1, column_value);
}

static void
set_column_value(Gtk::TreeModel::iterator &iter, Db::DataSet &row, int column_index)
{
  // NB: column_index starts from 1 (fricken stoopid DB people)
  
  Db::Type type( static_cast<Db::Type>(row->getMetaData()->getColumnType(column_index)) );
  
  switch(type)
    {
    case odbc::Types::VARCHAR:
    case odbc::Types::LONGVARCHAR:
      set_column_value_a<Glib::ustring, std::string>(iter, column_index, String(row->getString(column_index)) ); break;
    case odbc::Types::CHAR:
      set_column_value_a<char, char>(iter, column_index, row->getByte(column_index)); break;
    case odbc::Types::BIT:
      set_column_value_a<bool, bool>(iter, column_index, row->getBoolean(column_index)); break;
    case odbc::Types::INTEGER:
      set_column_value_a<int, int>(iter, column_index, row->getInt(column_index)); break;
    case odbc::Types::BIGINT:
      set_column_value_a<Db::Id, odbc::Long>(iter, column_index, row->getLong(column_index)); break;
    case odbc::Types::SMALLINT:
    case odbc::Types::TINYINT:
      set_column_value_a<short, short>(iter, column_index, row->getShort(column_index)); break;
    case odbc::Types::DECIMAL:
    case odbc::Types::NUMERIC:
      set_column_value_a<long double, double>(iter, column_index, row->getDouble(column_index)); break;
    case odbc::Types::DOUBLE:
    case odbc::Types::REAL:
      set_column_value_a<double, double>(iter, column_index, row->getDouble(column_index)); break;
    case odbc::Types::FLOAT:
      set_column_value_a<float, float>(iter, column_index, row->getFloat(column_index)); break;
    case odbc::Types::DATE:
      set_column_value_a<Db::Date, Db::Date>(iter, column_index, row->getDate(column_index)); break;
    case odbc::Types::TIMESTAMP:
      set_column_value_a<Db::DateTime, Db::DateTime>(iter, column_index, row->getTimestamp(column_index)); break;
    default:
      throw std::runtime_error(std::string("unknown data type (") + static_cast<std::string>(String(type)) + ") for data column");
    }
}



void
DataSelector:: append_row(DataSet &row)
{
  Gtk::TreeModel::iterator iter( _store->append() );
  unsigned col_count( row->getMetaData()->getColumnCount() );
  unsigned i;

  for (i = 1; i <= col_count; ++i)
    {
      WDEBUG_OUT << L"Setting column value for column '" << String(row->getMetaData()->getColumnName(i)) << L"'" << std::endl;
      set_column_value(iter, row, i);
      WDEBUG_OUT << L"...set column value for column '" << String(row->getMetaData()->getColumnName(i)) << L"'" << std::endl;
    }
}

StringList
DataSelector:: store_db_columns()
{
  StringList list;

  for (TreeModelColumnList::const_iterator iter(_store_columns.begin());
       iter != _store_columns.end();
       ++iter)
    {
      list.push_back(iter->first);
    }
  
  return list;
  
}


void
DataSelector:: refresh()
{
  Db::DataSetPtr data( _database->get_table_data(_table_name, store_db_columns()) );
  
  _store->clear();
  
  for ((*data)->next(); !(*data)->isAfterLast(); (*data)->next())
    append_row(*data);

  refreshed();
}

#define accessor(type, func, property) \
type \
DataSelector:: func() const \
{ \
  return property; \
}

accessor(const StringList &,                   get_selection_columns, _selection_columns);
accessor(const Glib::RefPtr<Gtk::ListStore> &, get_store,             _store);
accessor(const String &,                       get_table_name,        _table_name);
accessor(Db::Database *,                       get_database,          _database);
