/* -*- c++ -*-
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2008 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef __TK_EXPRESSION_H__
#define __TK_EXPRESSION_H__

#include "libtkcommon/tkstring.h"
#include "libtkcommon/numerical-types.h"
#include "libtkcommon/debug.h"

#include "libtkquery/query_tables.h"

namespace TkResolver {
  namespace Query {
    
    class QueryBuilder
      {
      public:
	virtual TkResolver::String build_query(QueryTables &tables) = 0;
      };
    
    
    class Value : public QueryBuilder
      {
      public:
	virtual TkResolver::String build_query(QueryTables &tables) = 0;
      };
    
    class Boolean : public Value
      {
      private:
	bool _val;
      public:
	Boolean (bool b)
	  : _val(b) {}
	
	virtual TkResolver::String build_query(QueryTables &tables);
      };
    
    class Integer : public Value
      {
      private:
	int _val;
      public:
	Integer (const char *text)
	  : _val(atoi(text)) {}
	
	virtual TkResolver::String build_query(QueryTables &tables);
      };
    
    class Real : public Value
      {
      private:
	double _val;
      public:
	Real (const char *text)
	  : _val(strtod(text, NULL)) {}
	
	virtual TkResolver::String build_query(QueryTables &tables);
      };
    
    class String : public Value
      {
      private:
	TkResolver::String _val;
      public:
	String (const TkResolver::String &str)
	{
	  // trim quote marks
	  TkResolver::String::const_iterator c( str.begin() );
	  for (++c; c != str.end();)
	    {
	      TkResolver::String::value_type chr( *c );
	      
	      if (++c != str.end())
		_val.push_back(chr);
	    }
	}
	
	virtual TkResolver::String build_query(QueryTables &tables);
      };
    
    class Identifier : public Value
      {
      private:
	TkResolver::String _name;
      public:
	Identifier (const TkResolver::String &name)
	{
	  // trim quote marks
	  TkResolver::String::const_iterator c( name.begin() );
	  for (++c; c != name.end();)
	    {
	      TkResolver::String::value_type chr( *c );
	      
	      if (++c != name.end())
		_name.push_back(chr);
	    }
	}
	
	virtual TkResolver::String build_query(QueryTables &tables);
      };
    
    
    class Op
      {
      private:
	TkResolver::String _op;
      public:
	Op (const char *text)
	  : _op(text) {}
	
	TkResolver::String get_op() const { return _op; }
      };
    
    
    class Expression : public QueryBuilder
      {
      public:
	virtual TkResolver::String build_query(QueryTables &tables) = 0;
      };
    
    class Clause : public Expression
      {
      public:
	virtual TkResolver::String build_query(QueryTables &tables) = 0;
      };
    
    class Binary : public Clause
      {
      private:
	Value *_val1;
	Op *_op;
	Value *_val2;
      public:
	Binary (Value *val1, Op *op, Value *val2)
	  : _val1(val1), _op(op), _val2(val2) {}
	~Binary()
	{ delete _val1; delete _op; delete _val2; }
	
	virtual TkResolver::String build_query(QueryTables &tables);
      };

    class Unary : public Clause
      {
      private:
	Op *_op;
	Value *_val;
      public:
	Unary (Op *op, Value *val)
	  : _op(op), _val(val) {}
	~Unary()
	{ delete _op; delete _val; }
	
	virtual TkResolver::String build_query(QueryTables &tables);
      };
    
    
    class Logical : public Expression
      {
      private:
	Expression *_expr1;
	Op *_lop;
	Expression *_expr2;
      public:
	Logical (Expression *expr1, Op *lop, Expression *expr2)
	  : _expr1(expr1), _lop(lop), _expr2(expr2) {}
	~Logical()
	{ delete _expr1; delete _lop; delete _expr2; }
	
	virtual TkResolver::String build_query(QueryTables &tables);
      };

    class Bracketed : public Expression
      {
      private:
	Expression *_expr;
      public:
	Bracketed (Expression *expr)
	  : _expr(expr) {}
	~Bracketed()
	{ delete _expr; }
	
	virtual TkResolver::String build_query(QueryTables &tables);
      };
    
    
  }; /* namespace Query */
}; /* namespace TkResolver */

#endif /* __TK_EXPRESSION_H__ */
