/*
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2004-2007 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <wchar.h>

#include <iostream>
#include <string>

#include <SDL/SDL.h>

#include "libtkcommon/exceptions.h"
#include "libtkcommon/debug.h"
#include "libtkcommon/real_time.h"
#include "libtkrandom/random_types.h"
#include "tkresolver.h"
#include "types.h"

using namespace TkResolver;

TKResolver:: TKResolver(const String &retriever_type,
			const String &algo,
			unsigned bit_multiplier,
			unsigned trace_count,
			unsigned total_time,
			unsigned width,
			unsigned height,
			bool scrolled,
			double speed,
			double horizon,
			bool test)
  : _trace_count(trace_count),
    _total_time(total_time),
    _display(new Display (height, width, _trace_count, _total_time, scrolled, speed, horizon)),
    _algorithm(algo),
    _decision_tree(_algorithm.get_outcome_count(), bit_multiplier),
    _test(test),
    _rand(retriever_type,
	  _decision_tree.deterministic(),
	  ((_display->get_total_lines() * _trace_count * bit_multiplier) / Random::BITS_PER_CHAR) + 1,
	  true),
    _current_line( _display->calc_start_line() )
{
}

TKResolver::~TKResolver()
{
  delete _display;
}

unsigned
TKResolver:: get_display_height() const
{
  return _display->get_height();
}

unsigned
TKResolver:: get_display_width() const
{
  return _display->get_width();
}

unsigned
TKResolver:: get_total_lines() const
{
  return _display->get_total_lines();
}

const Random::Bits&
TKResolver:: get_used_bits() const
{
  return _rand.get_used_bits();
}

bool
TKResolver:: control()
{
  SDL_Event event;

  while (SDL_PollEvent (&event))
    {
      switch (event.type)
	{
	case SDL_QUIT:
	  return true;

	case SDL_KEYUP:
	  switch (event.key.keysym.sym)
	    {
	    case SDLK_ESCAPE:
	      if (_test)
		return true;
	      break;

	    default:
	      break;
	    }
	  
	default:
	  break;
	}
      
    }

  return false;
}


void
TKResolver:: slice()
{
  Line line;
  int offset;
  for (unsigned short i = 0; i < _trace_count; ++i)
    {
      offset = _algorithm.choose_offset( _decision_tree.generate_event(_rand) );
      line.push_back(offset);
      _current_line[i] += offset;
    }
  
  _display->print_line(_current_line);
  
  _offsets.push_back(line);
}

static Result null_result;

const Result&
TKResolver:: run()
{
  thread_set_real_time(80);
  
  _display->count_down(5);
  
  tk_time_t slice_size;
  tk_time_t slice_start;
  long      wait_size;
  int       i;
  bool      user_quit;
  long      timing_error(0);
  
  slice_size = (_total_time * MILLISECONDS_PER_SECOND) / _display->get_total_lines();
  
  for (i = 0; i < _display->get_total_lines(); i++)
    {
      slice_start = SDL_GetTicks();
      
      user_quit = control();
      if (user_quit)
	return null_result;
      
      slice ();
      
      wait_size = slice_size - (SDL_GetTicks() - slice_start);
      if (wait_size > 0)
	{
	  SDL_Delay(wait_size);
	}
      else
	{
	  timing_error += -wait_size;
	  if (timing_error > 10000) // milliseconds
	    throw std::runtime_error("display is too slow");
	}
    }
  
  return _offsets;
}
